package net.sf.twip.parameterhandler;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;

import net.sf.twip.Assume;
import net.sf.twip.NotNull;
import net.sf.twip.Values;
import net.sf.twip.internal.TwipConfigurationError;
import net.sf.twip.util.Parameter;

/**
 * Produces all the values to call for one {@link Parameter}, including only
 * those that match the {@link Assume} or {@link NotNull} annotations.
 * 
 * TODO use a ServiceLoader for the subclasses of this class. There must be only
 * one for a type, but adding a subclass overloads the superclass (i.e. you
 * can't add more than one subclass).
 */
public abstract class ParameterHandler {

	protected final Parameter parameter;

	public ParameterHandler(Parameter parameter) {
		this.parameter = parameter;
	}

	/**
	 * Get the values from the method or field specified in the {@link Values}
	 * annotation... with a lot of collecting useful error information.
	 */
	private Object[] getAnnotatedValues(Parameter parameter) {
		final String name = parameter.getAnnotation(Values.class).value();
		try {
			final Field field = parameter.getMethod().getDeclaringClass().getField(name);
			if (!Modifier.isStatic(field.getModifiers()))
				throw new TwipConfigurationError("the field '" + name
						+ "' named in the @Values annotation of " + parameter + " must be static.");
			if (field.getType().isArray()) {
				if (field.getType().getComponentType().isPrimitive())
					throw new TwipConfigurationError("the values of the field '" + name
							+ "' named in the @Values annotation of " + parameter
							+ " must not be primitives.");
				return (Object[]) field.get(null);
			} else if (List.class.isAssignableFrom(field.getType())) {
				return ((List<?>) field.get(null)).toArray();
			} else {
				throw new TwipConfigurationError("the field '" + name
						+ "' named in the @Values annotation of " + parameter + " must be an array or list.");
			}
		} catch (final NoSuchFieldException ex) {
			try {
				final Method method = parameter.getMethod().getDeclaringClass().getMethod(name);
				if (!Modifier.isStatic(method.getModifiers()))
					throw new TwipConfigurationError("the method '" + name
							+ "' named in the @Values annotation of " + parameter + " must be static.");
				if (method.getReturnType().isArray()) {
					if (method.getReturnType().getComponentType().isPrimitive())
						throw new TwipConfigurationError("the values returned by the method '" + name
								+ "' named in the @Values annotation of " + parameter
								+ " must not be primitives.");
					return (Object[]) method.invoke(null);
				} else if (List.class.isAssignableFrom(method.getReturnType())) {
					return ((List<?>) method.invoke(null)).toArray();
				} else {
					throw new TwipConfigurationError("the method '" + name
							+ "' named in the @Values annotation of " + parameter
							+ " must return an array or list.");
				}
			} catch (final IllegalAccessException e) {
				throw new TwipConfigurationError("can't access the method '" + name
						+ "' named in the @Values annotation of " + parameter, e);
			} catch (final NoSuchMethodException e) {
				throw new TwipConfigurationError("there is no method or field '" + name
						+ "' named in the @Values annotation of " + parameter);
			} catch (final InvocationTargetException e) {
				throw new TwipConfigurationError("can't invoke method '" + name
						+ "' named in the @Values annotation of " + parameter, e);
			}
		} catch (final IllegalAccessException e) {
			throw new TwipConfigurationError("can't access field '" + name
					+ "' named in the @Values annotation of " + parameter, e);
		}
	}

	protected abstract Object[] getDefaultParameterValues();

	public Object[] getParameterValues() {
		if (parameter.isAnnotationPresent(Values.class)) {
			if (parameter.isAnnotationPresent(Assume.class))
				throw new TwipConfigurationError("don't annotate with Values AND with Assume");
			return getAnnotatedValues(parameter);
		}
		final Object[] defaultParameterValues = getDefaultParameterValues();
		if (!parameter.isAnnotationPresent(Assume.class))
			return defaultParameterValues;
		final List<Object> result = new ArrayList<Object>();
		for (final Object value : defaultParameterValues) {
			if (test(value)) {
				result.add(value);
			}
		}
		if (!parameter.isAnnotationPresent(NotNull.class) && !parameter.getType().isPrimitive()) {
			result.add(null);
		}
		return result.toArray();
	}

	/**
	 * Should this value go into the parameter values?
	 * 
	 * @param value
	 *            the value to check
	 */
	protected boolean test(Object value) {
		return true;
	}
}
